#!/usr/bin/env python
#!/usr/bin/python3
########################################################################################
# clover.py - Primary entry point for the CLOVER package.                              #
#                                                                                      #
# Author: Ben Winchester                                                               #
# Copyright: Ben Winchester, 2021                                                      #
# Date created: 16/09/2021                                                             #
# License: Open source                                                                 #
########################################################################################
"""
clover.py - Primary entry point for the CLOVER package.

CLOVER (Continuous Lifetime Optimisation of Variable Electricity Resources) can evaluate
and optimise minigrid systems, determining whether a demand is met whilst minimising
environmental and economic impacts. The main flow of CLOVER can be executed by running
the clover module from the command-line interface.

"""

import sys

from ..__main__ import main as clover_main


def main() -> None:
    """
    Main function of the CLOVER entry-point script.

    """

    clover_main(sys.argv[1:])



#!/usr/bin/env python
########################################################################################
# clover_update_api_token.py - Entry point for updating renewable.ninja api tokens.    #
#                                                                                      #
# Author: Ben Winchester                                                               #
# Copyright: Ben Winchester, 2021                                                      #
# Date created: 21/09/2021                                                             #
# License: Open source                                                                 #
########################################################################################
"""
clover_update_api_token.py - Entry point for updating location API tokens.

CLOVER works on the basis of locations. The solar, wind and weather data for these
locations is fetched from the freely-accessible online renewables.ninja resource. In
order to properly download these data, each user must specify a unique API, which must
be entered as an input to CLOVER in order for it to properly function.

"""

import sys

from .update_api_token import main as update_api_token_main


def main() -> None:
    """
    Main function of the update-api-token entry-point script.

    """

    update_api_token_main(sys.argv[1:])



#!/usr/bin/python3
########################################################################################
# conversion.py - Conversion module for CLOVER.                                        #
#                                                                                      #
# Authors: Phil Sandwell, Ben Winchester                                               #
# Copyright: Phil Sandwell, 2021                                                       #
# License: Open source                                                                 #
# Most recent update: 14/07/2021                                                       #
#                                                                                      #
# For more information, please email:                                                  #
#     philip.sandwell@gmail.com                                                        #
########################################################################################
"""
conversion.py - The conversion module of CLOVER.

This module contains functionality enabling the conversion of one energy type into
another.

"""

from logging import Logger
from typing import Any, Dict, List, Optional

from ..__utils__ import (
    BColours,
    HTFMode,
    InputFileError,
    NAME,
    RESOURCE_NAME_TO_RESOURCE_TYPE_MAPPING,
    ResourceType,
)
from ..impact.__utils__ import WasteProduct

__all__ = (
    "Converter",
    "MultiInputConverter",
    "ThermalDesalinationPlant",
    "WaterSource",
)


# Heat source:
#   Keyword used for parsing the heat source of thermal desalination plants.
HEAT_SOURCE: str = "heat_source"

# Maximum feedwater input temperature:
#   Keyword used for parsing maximum feedwater temperature information.
MAXIMUM_FEEDWATER_TEMPERATURE: str = "max_feedwater_temperature"

# Maximum htf input temperature:
#   Keyword used for parsing maximum htf temperature information.
MAXIMUM_HTF_TEMPERATURE: str = "max_htf_temperature"

# Maximum output:
#   Keyword used for parsing maximum output information.
MAXIMUM_OUTPUT: str = "maximum_output"

# Maximum feedwater input temperature:
#   Keyword used for parsing maximum feedwater temperature information.
MINIMUM_FEEDWATER_TEMPERATURE: str = "min_feedwater_temperature"

# Minimum water input temperature:
#   Keyword used for parsing minimum htf temperature information.
MINIMUM_HTF_TEMPERATURE: str = "min_htf_temperature"

# Minimum output:
#   Keyword used for parsing minimum output information.
MINIMUM_OUTPUT: str = "minimum_output"

# Output:
#   Keyword used for parsing output information.
OUTPUT: str = "output"

# Waste products:
#   Keyword used for parsing waste-product information.
WASTE_PRODUCTS: str = "waste_products"


def _parse_waste_production(
    logger: Logger, name: str, waste_product_inputs: Dict[str, float]
) -> Dict[WasteProduct, float]:
    """
    Parses waste-product information.

    Inputs:
        - logger:
            The :class:`logging.Logger` to use for the run.
        - name:
            The name of the :class:`Converter` being parsed.
        - waste_product_inputs:
            The waste-product input information for the :class:`Converter` being parsed.

    Outputs:
        - A mapping between the :class:`WasteProduct` and its associated output
        produced.

    """

    waste_production: Dict[WasteProduct, float] = {}

    for waste_product, amount_produced in waste_product_inputs.items():
        try:
            waste_production[WasteProduct(waste_product)] = amount_produced
        except ValueError:
            logger.error(
                "%sInvalid waste product specified: '%s'. Valid values are %s.%s",
                BColours.fail,
                waste_product,
                ", ".join(e.value for e in WasteProduct),
                BColours.endc,
            )
            raise InputFileError(
                "conversion inputs",
                f"Converter {name} has invalid waste-product '{waste_product}'. Valid "
                + f"waste products are {', '.join(str(e.value) for e in WasteProduct)}.",
            ) from None

        # Type check the value generated.
        if not isinstance(waste_production[WasteProduct(waste_product)], (int, float)):
            logger.error(
                "%sInvalid value for waste-product '%s' for plant '%s': "
                "value must be of type float.%s",
                BColours.fail,
                waste_product,
                BColours.endc,
            )
            raise InputFileError(
                "conversion inputs",
                f"Converter {name} has invalid waste-product type.",
            )

    return waste_production


class Converter:
    """
    Represents a device that is able to convert one from of energy into another.

    .. attribute:: consumption
        The amount of input load which is consumed per unit output load produced.

    .. attribute:: input_resource_consumption
        A mapping between :class:`ResourceType` and the amount of input required, from
        that :class:`ResourceType`, when the :class:`Converter` is operating at its
        maximum throughput.

    .. attribute:: maximum_output_capacity
        The maximum capacity of the device in producing its output.

    .. attribute:: name
        The name of the :class:`Converter` instance.

    .. attribute:: output_resource_type
        The type of energy which is outputted by the device.

    .. attribute:: waste_production
        A mapping between :class:`WasteProduct` instances and the amount of each waste
        product that is produced per unit output.

    """

    def __init__(
        self,
        input_resource_consumption: Dict[ResourceType, float],
        maximum_output_capacity: float,
        name: str,
        output_resource_type: ResourceType,
        waste_production: Optional[Dict[WasteProduct, float]] = None,
    ) -> None:
        """
        Instantiate a :class:`Converter` instance.

        Inputs:
            - input_resource_types:
                The types of load inputted to the converter.
            - maximum_output_capcity:
                The maximum output capacity of the converter.
            - name:
                The name of the converter.
            - output_resource_type:
                The type of output produced by the converter.
            - waste_production:
                The waste production of the converter.

        """

        self.input_resource_consumption: Dict[
            ResourceType, float
        ] = input_resource_consumption
        self.maximum_output_capacity: float = maximum_output_capacity
        self.name: str = name
        self.output_resource_type: ResourceType = output_resource_type
        self.waste_production: Dict[WasteProduct, float] = (
            waste_production if waste_production is not None else {}
        )

    def __eq__(self, other: Any) -> bool:
        """
        Returns whether two :class:`Conversion` instances are equal.

        The comparison is made on the consumption and whether their input and output
        types are the same.

        Outputs:
            - Whether the two instances are equal.

        """

        return bool(
            self.input_resource_consumption == other.input_resource_consumption
            and self.output_resource_type == other.output_resource_type
            and self.consumption == other.consumption
            and self.waste_production == other.waste_production
        )

    def __hash__(self) -> int:
        """
        Returns a unique `int` identifying the :class:`Converter` instance.

        Outputs:
            A unique `int` identifying the :class:`Converter` instance.

        """

        consumption_dict = dict(enumerate(self.input_resource_consumption.values()))
        consumption_int = sum(
            value * 10 ** (5 * index) for index, value in consumption_dict.items()
        )

        return hash((consumption_int + self.maximum_output_capacity) ** 2)

    def __lt__(self, other: Any) -> bool:
        """
        Returns whether the current instance is less than another instance.

        The comparison is made purely on the consumption.

        Outputs:
            - Whether the current instance is less than the other in consumption.

        Raises:
            - An Exception if the two instances are not of the same type and an attempt
              was made at this comparison.

        """

        if self.output_resource_type != other.output_resource_type:
            raise Exception(
                "An attempt was made to compare two conversion instances that use "
                "different output types."
            )

        return bool(self.consumption < other.consumption)

    def __repr__(self) -> str:
        """
        Returns a nice-looking `str` representing the :class:`Converter` instance.

        Outputs:
            - A nidee-looking `str` representing the :class:`Converter` instance.

        """

        return self.__str__()

    def __str__(self) -> str:
        """
        Returns a nice-looking `str` representing the :class:`Converter` instance.

        Outputs:
            - A nidee-looking `str` representing the :class:`Converter` instance.

        """

        input_resource_consumption = ", ".join(
            [
                f"{key.value}={value}"
                for key, value in self.input_resource_consumption.items()
            ]
        )

        return (
            "Converter("
            + f"name={self.name}"
            + f", input_resource_consumption=({input_resource_consumption})"
            + f", output_resource_type = {self.output_resource_type.value}"
            + f", maximum_output_capacity = {self.maximum_output_capacity}"
            + (
                f", waste_production = {self.waste_production}"
                if len(self.waste_production) > 0
                else ""
            )
            + ")"
        )

    @property
    def consumption(self) -> float:
        """
        Used only when dealing with a single input resource type.

        Outputs:
            - The consumption of the input resource type.

        """

        if len(self.input_resource_consumption) > 1:
            raise InputFileError(
                "conversion inputs",
                "Multiple inputs were defined where only one was expected on a "
                + f"converter instance: {self.name}",
            )

        return list(self.input_resource_consumption.values())[0]

    @property
    def value(self) -> str:
        """
        Used to mimic the behaviour of enums.

        In order to utilise :class:`Converter` instances in the same way that
        :class:`enum.Enum` instances are, it is necessary to include a '.value'
        property.

        Outputs:
            - A `str` giving the value associated with the :class:`Converter` instance,
              i.e., its name.

        """

        return self.name


class MultiInputConverter(Converter):
    """
    Represents a converter that is capable of having multiple input resource types.

    """

    def __lt__(self, other: Any) -> bool:
        """
        Returns whether the current instance is less than another instance.

        The comparison is made purely on the consumption.

        Outputs:
            - Whether the current instance is less than the other in consumption.

        Raises:
            - An Exception if the two instances are not of the same type and an attempt
              was made at this comparison.

        """

        return bool(
            list(self.input_resource_consumption.values())[0]
            < list(other.input_resource_consumption.values())[0]
        )

    @classmethod
    def from_dict(cls, input_data: Dict[str, Any], logger: Logger) -> Any:
        """
        Generates a :class:`MultiInputConverter` instance based on the input data.

        Inputs:
            - input_data:
                The input data, parsed from the input file.

        Outputs:
            - A :class:`MultiInputConverter` instance based on the input data.

        """

        if not all(isinstance(key, str) for key in input_data):
            raise InputFileError(
                "conversion inputs", "All conversion input keys must be of type `str`."
            )

        # Determine the input load type.
        input_resource_list: List[str] = [
            str(key)
            for key in input_data
            if key in RESOURCE_NAME_TO_RESOURCE_TYPE_MAPPING
        ]
        # Determine the output load type.
        try:
            output_resource_type = ResourceType(input_data[OUTPUT])
        except KeyError as e:
            logger.error(
                "%sOutput load type of water pump is not valid: %s%s",
                BColours.fail,
                str(e),
                BColours.endc,
            )
            raise Exception(
                f"{BColours.fail}Output load type invalid: {str(e)}{BColours.endc}"
            ) from None

        # Determine the power consumption of the device.
        maximum_output = input_data[MAXIMUM_OUTPUT]
        try:
            maximum_output = float(maximum_output)
        except TypeError as e:
            logger.error(
                "%sInvalid entry in conversion file, check all value types are "
                "correct: %s%s",
                BColours.fail,
                str(e),
                BColours.endc,
            )
            raise Exception(
                f"{BColours.fail}Invalid value type in conversion file: {str(e)}{BColours.endc}"
            ) from None

        input_resource_consumption: Dict[ResourceType, float] = {}

        for input_resource in input_resource_list:
            try:
                input_resource_consumption[
                    ResourceType(RESOURCE_NAME_TO_RESOURCE_TYPE_MAPPING[input_resource])
                ] = float(input_data[input_resource])
            except TypeError as e:
                logger.error(
                    "%sInvalid entry in conversion file, check all value types are "
                    "correct: %s%s",
                    BColours.fail,
                    str(e),
                    BColours.endc,
                )
                raise InputFileError(
                    "conversion inputs",
                    f"Invalid value type in conversion file: {str(e)}",
                ) from None

        waste_production: Dict[WasteProduct, float] = {}

        if WASTE_PRODUCTS in input_data:
            waste_production = _parse_waste_production(
                logger, input_data[NAME], input_data[WASTE_PRODUCTS]
            )

        return cls(
            input_resource_consumption,
            maximum_output,
            str(input_data[NAME]),
            output_resource_type,
            waste_production,
        )


class ThermalDesalinationPlant(MultiInputConverter):
    """
    Represents a thermal desalination plant.

    .. attribute:: htf_mode
        The mode of inputting HTF to the thermal desalination plant.

    .. attribute:: maximum_feedwater_temperature
        The maximum temperature of feedwater allowed by the plant, measured in degrees
        Celcius.

    .. attribute:: maximum_htf_temperature
        The maximum temperature of HTF allowed by the plant, measured in degrees
        Celcius.

    .. attribute:: minimum_feedwater_temperature
        The minimum temperature of feedwater allowed by the plant, measured in degrees
        Celcius.

    .. attribute:: minimum_htf_temperature
        The minumum temperature of HTF allowed by the plant, measured in degrees
        Celcius.

    .. attribute:: minimum_output_capacity
        The minimum output flow rate of the plant.

    """

    def __init__(
        self,
        htf_mode: HTFMode,
        input_resource_consumption: Dict[ResourceType, float],
        maximum_feedwater_temperature: Optional[float],
        maximum_htf_temperature: Optional[float],
        maximum_output_capacity: float,
        minimum_feedwater_temperature: Optional[float],
        minimum_htf_temperature: Optional[float],
        minimum_output_capacity: float,
        name: str,
        output_resource_type: ResourceType,
        waste_production: Dict[WasteProduct, float],
    ) -> None:
        """
        Instantiate a :class:`ThermalDesalinationPlant` instance.

        Inputs:
            - htf_mode:
                The mode of inputting heat to the plant.
            - input_resource_types:
                The types of load inputted to the plant.
            - maximum_feedwater_temperature:
                The maximum temperature of feedwater allowed into the plant, measured in
                degrees Celcius.
            - maximum_htf_temperature:
                The maximum temperature of water allowed into the plant, measured in
                degrees Celcius.
            - maximum_output_capcity:
                The maximum output capacity of the plant.
            - minimum_feedwater_temperature:
                The minimum temperature of feedwater allowed into the plant, measured in
                degrees Celcius.
            - minimum_htf_temperature:
                The mibimum temperature of water allowed into the plant, measured in
                degrees Celcius.
            - minimum_output_capcity:
                The minimum output capacity of the plant.
            - name:
                The name of the plant.
            - output_resource_type:
                The type of output produced by the plant.
            - waste_production:
                The waste products produced.

        """

        super().__init__(
            input_resource_consumption,
            maximum_output_capacity,
            name,
            output_resource_type,
            waste_production,
        )

        self.htf_mode = htf_mode
        self.maximum_feedwater_temperature: Optional[
            float
        ] = maximum_feedwater_temperature
        self.maximum_htf_temperature: Optional[float] = maximum_htf_temperature
        self.minimum_feedwater_temperature: Optional[
            float
        ] = minimum_feedwater_temperature
        self.minimum_htf_temperature: Optional[float] = minimum_htf_temperature
        self.minimum_output_capacity: Optional[float] = minimum_output_capacity

    @classmethod
    def from_dict(cls, input_data: Dict[str, Any], logger: Logger) -> Any:
        """
        Generates a :class:`ThermalDesalinationPlant` instance based on the input data.

        Inputs:
            - input_data:
                The input data, parsed from the input file.

        Outputs:
            - A :class:`ThermalDesalinationPlant` instance based on the input data.

        """

        if not all(isinstance(key, str) for key in input_data):
            raise InputFileError(
                "conversion inputs", "All conversion input keys must be of type `str`."
            )

        # Determine the input load type.
        input_resource_list: List[str] = [
            str(key)
            for key in input_data
            if key in RESOURCE_NAME_TO_RESOURCE_TYPE_MAPPING
        ]
        # Determine the output load type.
        try:
            output_resource_type = ResourceType(input_data[OUTPUT])
        except KeyError as e:
            logger.error(
                "%sOutput load type of water pump is not valid: %s%s",
                BColours.fail,
                str(e),
                BColours.endc,
            )
            raise Exception(
                f"{BColours.fail}Output load type invalid: {str(e)}{BColours.endc}"
            ) from None

        # Determine the power consumption of the device.
        maximum_output = input_data[MAXIMUM_OUTPUT]
        try:
            maximum_output = float(maximum_output)
        except TypeError as e:
            logger.error(
                "%sInvalid entry in conversion file, check all value types are "
                "correct: %s%s",
                BColours.fail,
                str(e),
                BColours.endc,
            )
            raise Exception(
                f"{BColours.fail}Invalid value type in conversion file: {str(e)}{BColours.endc}"
            ) from None

        input_resource_consumption: Dict[ResourceType, float] = {}

        for input_resource in input_resource_list:
            try:
                input_resource_consumption[
                    ResourceType(RESOURCE_NAME_TO_RESOURCE_TYPE_MAPPING[input_resource])
                ] = float(input_data[input_resource])
            except TypeError as e:
                logger.error(
                    "%sInvalid entry in conversion file, check all value types are "
                    "correct: %s%s",
                    BColours.fail,
                    str(e),
                    BColours.endc,
                )
                raise Exception(
                    f"{BColours.fail}Invalid value type in conversion file: {str(e)}{BColours.endc}"
                ) from None

        try:
            htf_mode = HTFMode(input_data[HEAT_SOURCE])
        except KeyError:
            logger.info(
                "%sThe plant, %s, did not specify the source of heat. Cannot "
                "create a thermal desalination plant for %s%s",
                BColours.fail,
                str(input_data[NAME]),
                str(input_data[NAME]),
                BColours.endc,
            )
            raise

        waste_production: Dict[WasteProduct, float] = {}

        if WASTE_PRODUCTS in input_data:
            waste_production = _parse_waste_production(
                logger, input_data[NAME], input_data[WASTE_PRODUCTS]
            )

        return cls(
            htf_mode,
            input_resource_consumption,
            float(input_data[MAXIMUM_FEEDWATER_TEMPERATURE])
            if MAXIMUM_FEEDWATER_TEMPERATURE in input_data
            else None,
            float(input_data[MAXIMUM_HTF_TEMPERATURE])
            if MAXIMUM_HTF_TEMPERATURE in input_data
            else None,
            maximum_output,
            float(input_data[MINIMUM_FEEDWATER_TEMPERATURE])
            if MINIMUM_FEEDWATER_TEMPERATURE in input_data
            else None,
            float(input_data[MINIMUM_HTF_TEMPERATURE])
            if MINIMUM_HTF_TEMPERATURE in input_data
            else None,
            float(input_data[MINIMUM_OUTPUT]),
            str(input_data[NAME]),
            output_resource_type,
            waste_production,
        )


class WaterSource(Converter):
    """Represents a water source which takes in electricity and outputs water."""

    def __str__(self) -> str:
        """
        Returns a nice-looking `str` representing the :class:`Converter` instance.

        Outputs:
            - A nidee-looking `str` representing the :class:`Converter` instance.

        """

        input_resource_consumption = ", ".join(
            [
                f"{key.value}={value} units/output unit"
                for key, value in self.input_resource_consumption.items()
            ]
        )

        return (
            "Converter("
            + f"name={self.name}"
            + f", input_resource_consumption=({input_resource_consumption})"
            + f", output_resource_type = {self.output_resource_type.value}"
            + f", maximum_output_capacity = {self.maximum_output_capacity}"
            + ")"
        )

    @classmethod
    def from_dict(cls, input_data: Dict[str, Any], logger: Logger) -> Any:
        """
        Generates a :class:`Converter` instance based on the input data provided.

        Inputs:
            - input_data:
                The input data, parsed from the input file.

        Outputs:
            - A :class:`Converter` instance based on the input data.

        """

        if not all(isinstance(key, str) for key in input_data):
            raise InputFileError(
                "conversion inputs", "All conversion input keys must be of type `str`."
            )

        # Determine the input load type.
        input_resource_list: List[str] = [
            str(key)
            for key in input_data
            if key in RESOURCE_NAME_TO_RESOURCE_TYPE_MAPPING
        ]
        if len(input_resource_list) > 1:
            logger.info(
                "%sCurrently only one input is allowed for water sources.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InputFileError(
                "conversion inputs",
                f"{BColours.fail}Currently only one load type is supported in the "
                + f"conversion inputs file.{BColours.endc}",
            )

        input_resource_type = ResourceType(
            RESOURCE_NAME_TO_RESOURCE_TYPE_MAPPING[input_resource_list[0]]
        )

        # Determine the output load type.
        try:
            output_resource_type = ResourceType(input_data[OUTPUT])
        except KeyError as e:
            logger.error(
                "%sOutput load type of water pump is not valid: %s%s",
                BColours.fail,
                str(e),
                BColours.endc,
            )
            raise Exception(
                f"{BColours.fail}Output load type invalid: {str(e)}{BColours.endc}"
            ) from None

        # Determine the power consumption of the device.
        maximum_output = input_data[MAXIMUM_OUTPUT]
        corresponding_input = input_data[input_resource_list[0]]
        try:
            maximum_output = float(maximum_output)
        except TypeError as e:
            logger.error(
                "%sInvalid entry in conversion file, check all value types are "
                "correct: %s%s",
                BColours.fail,
                str(e),
                BColours.endc,
            )
            raise Exception(
                f"{BColours.fail}Invalid value type in conversion file: {str(e)}{BColours.endc}"
            ) from None

        try:
            corresponding_input = float(corresponding_input)
        except TypeError as e:
            logger.error(
                "%sInvalid entry in conversion file, check all value types are "
                "correct: %s%s",
                BColours.fail,
                str(e),
                BColours.endc,
            )
            raise Exception(
                f"{BColours.fail}Invalid value type in conversion file: {str(e)}{BColours.endc}"
            ) from None

        consumption = corresponding_input / maximum_output

        waste_production: Dict[WasteProduct, float] = {}

        if WASTE_PRODUCTS in input_data:
            waste_production = _parse_waste_production(
                logger, input_data[NAME], input_data[WASTE_PRODUCTS]
            )

        return cls(
            {input_resource_type: consumption},
            maximum_output,
            str(input_data[NAME]),
            output_resource_type,
            waste_production,
        )



#!/usr/bin/env python
########################################################################################
# clover_hpc_clover.py - Entry point for running CLOVER on Imperial College's HPC.     #
#                                                                                      #
# Author: Ben Winchester                                                               #
# Copyright: Ben Winchester, 2021                                                      #
# Date created: 08/03/2022                                                             #
# License: Open source                                                                 #
########################################################################################
"""
clover_hpc_clover.py - Entry point for running CLOVER on the HPC.

CLOVER can be run on Imperial College London's high-performance computers (HPC)s. In
order to do so as an installed piece of software, this wrapper provides an entry point
for the code once installed.

"""

import sys

from .hpc_clover import main as hpc_clover_main


def main() -> None:
    """
    Main function of the CLOVER-HPC entry-point script.

    """

    hpc_clover_main(sys.argv[1:])



#!/usr/bin/python3
########################################################################################
# ghgs.py - GHG impact assessment module.                                              #
#                                                                                      #
# Author: Phil Sandwell, Ben Winchester                                                #
# Copyright: Phil Sandwell, 2021                                                       #
# License: Open source                                                                 #
# Most recent update: 24/08/2021                                                       #
#                                                                                      #
# For more information, please email:                                                  #
#     philip.sandwell@gmail.com                                                        #
########################################################################################
"""
ghgs.py - The ghgs module for CLOVER.

When assessing the impact of a system, the ghg impact, i.e., the greenhouse gasses
emitted by the system, need to be assed.

"""

from logging import Logger
from typing import Any, Dict, List, Optional

import numpy as np  # pylint: disable=import-error
import pandas as pd  # pylint: disable=import-error

from ..__utils__ import (
    BColours,
    ColumnHeader,
    hourly_profile_to_daily_sum,
    InputFileError,
    Location,
    Scenario,
)
from .__utils__ import ImpactingComponent, LIFETIME, SIZE_INCREMENT

__all__ = (
    "calculate_connections_ghgs",
    "calculate_diesel_fuel_ghgs",
    "calculate_grid_ghgs",
    "calculate_independent_ghgs",
    "calculate_kerosene_ghgs",
    "calculate_kerosene_ghgs_mitigated",
    "calculate_total_equipment_ghgs",
    "calculate_total_om",
    "EMISSIONS",
)


# Connection ghgs:
#   Keyword used for denoting connection ghgs.
CONNECTION_GHGS = "connection_ghgs"

# Emissions:
#   Keyword used for denoting device-specific ghg information.
EMISSIONS = "emissions"

# Extension ghgs:
#   Keyword used for denoting extension ghgs.
EXTENSION_GHGS = "extension_ghgs"

# GHGS:
#   Keyword for describing GHG data.
GHGS = "ghgs"

# GHG_DECREASE:
#   Keyword for describing GHG decrease.
GHG_DECREASE = "ghg_decrease"

# GHG impact:
#   A base `str` used for specifying unique ghg impacts.
GHG_IMPACT: str = "{type}_{name}"

# Initial GHGs:
#   Keyword for the initial GHGs.
INITIAL_GHGS = "initial_ghgs"

# Infrastructure Ghgs:
#   Keyword used for denoting infrastructure ghgs.
INFRASTRUCTURE_GHGS = "infrastructure_ghgs"

# Installation GHGS:
#   Keyword for describing GHG data.
INSTALLATION_GHGS = "installation_ghgs"

# GHG_DECREASE:
#   Keyword for describing GHG decrease.
INSTALLATION_GHGS_DECREASE = "installation_ghg_decrease"

# Final GHGs:
#   Keyword for the final GHGs.
FINAL_GHGS = "final_ghgs"

# OM GHGs:
#   Keyword for the O&M GHGs.
OM_GHGS = "o&m"


def calculate_ghgs(
    capacity: float,
    ghg_inputs: Dict[str, Any],
    system_component: str,
    year: int = 0,
) -> float:
    """
    Calculates ghgs of PV

    Inputs:
        - capacity:
            The capacity of the relevant component.
        - ghg_inputs:
            GHG input data.
        - system_component:
            The component of the system currently being considered, passed in as a `str`
            representing a component rather than a component.
        - year:
            ColumnHeader.INSTALLATION_YEAR.value

    Outputs:
        GHGs

    """

    ghgs: float = capacity * float(ghg_inputs[system_component][GHGS])
    annual_reduction: float = 0.01 * ghg_inputs[system_component][GHG_DECREASE]
    return ghgs * (1.0 - annual_reduction) ** year


# Installation ghgs
def calculate_installation_ghgs(
    capacity: float,
    ghg_inputs: Dict[str, Any],
    system_component: str,
    year: int = 0,
) -> float:
    """

    Calculates ghgs of installation
    Inputs:
        - capacity:
            The capacity of the relevant component.
        - ghg_inputs:
            GHG input data.
        - system_component:
            The component of the system currently being considered, passed in as a `str`
            representing the component rather than the component.
        - year:
            ColumnHeader.INSTALLATION_YEAR.value

    Outputs:
        GHGs

    """

    installation_ghgs: float = (
        capacity * ghg_inputs[system_component][INSTALLATION_GHGS]
    )
    annual_reduction: float = (
        0.01 * ghg_inputs[system_component][INSTALLATION_GHGS_DECREASE]
    )

    return installation_ghgs * (1.0 - annual_reduction) ** year


#   Miscellaneous ghgs
def calculate_misc_ghgs(capacity: float, ghg_inputs: Dict[str, Any]) -> float:
    """
    Calculates ghgs of miscellaneous capacity-related equipment

    Inputs:
        - capacity:
            The capacity of the relevant component.
        - ghg_inputs:
            GHG input data.

    Outputs:
        Misc. ghgs

    """

    misc_ghgs: float = capacity * ghg_inputs[ImpactingComponent.MISC.value][GHGS]
    return misc_ghgs


def calculate_total_equipment_ghgs(  # pylint: disable=too-many-locals
    buffer_tanks: int,
    clean_water_tanks: int,
    converters: Dict[str, int],
    diesel_size: float,
    ghg_inputs: Dict[str, Any],
    heat_exchangers: int,
    hot_water_tanks: int,
    logger: Logger,
    pv_array_size: float,
    pvt_array_size: float,
    storage_size: float,
    year: int = 0,
) -> float:
    """
    Calculates ghgs of all newly installed equipment

    Inputs:
        - buffer_tanks:
            Number of buffer tanks being installed.
        - clean_water_tanks:
            Capacity of clean-water tanks being installed.
        - converters:
            A mapping between converter names and the size of each that was added to the
            system this iteration.
        - diesel_size:
            Capacity of diesel generator being installed
        - ghg_inputs:
            GHG input information.
        - heat_exchangers:
            Number of heat exchangers being installed.
        - hot_water_tanks:
            Capactiy of hot-water tanks being installed.
        - pv_array_size:
            Capacity of PV being installed.
        - pvt_array_size:
            Capacity of PV-T being installed.
        - storage_size:
            Capacity of battery storage being installed.
        - year:
            ColumnHeader.INSTALLATION_YEAR.value.

    Outputs:
        GHGs

    """

    # Calculate system ghgs.
    bos_ghgs = calculate_ghgs(
        pv_array_size, ghg_inputs, ImpactingComponent.BOS.value, year
    )

    if ImpactingComponent.BUFFER_TANK.value not in ghg_inputs and buffer_tanks > 0:
        logger.error(
            "%sNo buffer-tank GHG input information provided.%s",
            BColours.fail,
            BColours.endc,
        )
        raise InputFileError(
            "tank inputs",
            "No buffer tank ghg input information provided and a non-zero number of "
            "buffer tanks are being considered.",
        )
    buffer_tank_ghgs: float = 0
    buffer_tank_installation_ghgs: float = 0
    if buffer_tanks > 0:
        buffer_tank_ghgs = calculate_ghgs(
            buffer_tanks,
            ghg_inputs,
            ImpactingComponent.BUFFER_TANK.value,
            year,
        )
        buffer_tank_installation_ghgs = calculate_installation_ghgs(
            buffer_tanks, ghg_inputs, ImpactingComponent.BUFFER_TANK.value, year
        )

    if (
        ImpactingComponent.CLEAN_WATER_TANK.value not in ghg_inputs
        and clean_water_tanks > 0
    ):
        logger.error(
            "%sNo clean-water tank GHG input information provided.%s",
            BColours.fail,
            BColours.endc,
        )
        raise InputFileError(
            "tank inputs",
            "No clean-water tank ghg input information provided and a non-zero number "
            "of clean-water tanks are being considered.",
        )
    clean_water_tank_ghgs: float = 0
    clean_water_tank_installation_ghgs: float = 0
    if clean_water_tanks > 0:
        clean_water_tank_ghgs = calculate_ghgs(
            clean_water_tanks,
            ghg_inputs,
            ImpactingComponent.CLEAN_WATER_TANK.value,
            year,
        )
        clean_water_tank_installation_ghgs = calculate_installation_ghgs(
            clean_water_tanks,
            ghg_inputs,
            ImpactingComponent.CLEAN_WATER_TANK.value,
            year,
        )

    converter_ghgs = sum(
        calculate_ghgs(
            size,
            ghg_inputs,
            GHG_IMPACT.format(type=ImpactingComponent.CONVERTER.value, name=converter),
            year,
        )
        for converter, size in converters.items()
    )
    converter_installation_ghgs = sum(
        calculate_installation_ghgs(
            size,
            ghg_inputs,
            GHG_IMPACT.format(type=ImpactingComponent.CONVERTER.value, name=converter),
            year,
        )
        for converter, size in converters.items()
    )

    diesel_ghgs = calculate_ghgs(
        diesel_size, ghg_inputs, ImpactingComponent.DIESEL.value, year
    )
    diesel_installation_ghgs = calculate_installation_ghgs(
        diesel_size, ghg_inputs, ImpactingComponent.DIESEL.value, year
    )

    if (
        ImpactingComponent.HEAT_EXCHANGER.value not in ghg_inputs
        and heat_exchangers > 0
    ):
        logger.error(
            "%sNo heat-exchanger tank GHG input information provided.%s",
            BColours.fail,
            BColours.endc,
        )
        raise InputFileError(
            "heat exchanger inputs",
            "No heat-exchanger ghg input information provided and a non-zero number of "
            "heat exchangers are being considered.",
        )
    heat_exchanger_ghgs: float = 0
    heat_exchanger_installation_ghgs: float = 0
    if heat_exchangers > 0:
        heat_exchanger_ghgs = calculate_ghgs(
            heat_exchangers,
            ghg_inputs,
            ImpactingComponent.HEAT_EXCHANGER.value,
            year,
        )
        heat_exchanger_installation_ghgs = calculate_installation_ghgs(
            heat_exchangers, ghg_inputs, ImpactingComponent.HEAT_EXCHANGER.value, year
        )

    if (
        ImpactingComponent.HOT_WATER_TANK.value not in ghg_inputs
        and hot_water_tanks > 0
    ):
        logger.error(
            "%sNo hot-water tank GHG input information provided.%s",
            BColours.fail,
            BColours.endc,
        )
        raise InputFileError(
            "tank inputs",
            "No hot-water tank ghg input information provided and a non-zero number of "
            "hot-water tanks are being considered.",
        )
    hot_water_tank_ghgs: float = 0
    hot_water_tank_installation_ghgs: float = 0
    if hot_water_tanks > 0:
        hot_water_tank_ghgs = calculate_ghgs(
            hot_water_tanks,
            ghg_inputs,
            ImpactingComponent.HOT_WATER_TANK.value,
            year,
        )
        hot_water_tank_installation_ghgs = calculate_installation_ghgs(
            hot_water_tanks, ghg_inputs, ImpactingComponent.HOT_WATER_TANK.value, year
        )

    pv_ghgs = calculate_ghgs(
        pv_array_size, ghg_inputs, ImpactingComponent.PV.value, year
    )
    pv_installation_ghgs = calculate_installation_ghgs(
        pv_array_size, ghg_inputs, ImpactingComponent.PV.value, year
    )

    if ImpactingComponent.PV_T.value not in ghg_inputs and pvt_array_size > 0:
        logger.error(
            "%sNo PV-T GHG input information provided.%s", BColours.fail, BColours.endc
        )
        raise InputFileError(
            "solar generation inputs",
            "No PV-T ghg input information provided and a non-zero number of PV-T"
            "panels are being considered.",
        )
    pvt_ghgs: float = 0
    pvt_installation_ghgs: float = 0
    if pvt_array_size > 0:
        pvt_ghgs = calculate_ghgs(
            pvt_array_size,
            ghg_inputs,
            ImpactingComponent.PV_T.value,
            year,
        )
        pvt_installation_ghgs = calculate_installation_ghgs(
            pvt_array_size, ghg_inputs, ImpactingComponent.PV.value, year
        )

    storage_ghgs = calculate_ghgs(
        storage_size, ghg_inputs, ImpactingComponent.STORAGE.value, year
    )

    # Calculate misc GHGs.
    misc_ghgs = calculate_misc_ghgs(diesel_size + pv_array_size, ghg_inputs)

    return (
        bos_ghgs
        + buffer_tank_ghgs
        + buffer_tank_installation_ghgs
        + clean_water_tank_installation_ghgs
        + clean_water_tank_ghgs
        + converter_ghgs
        + converter_installation_ghgs
        + diesel_installation_ghgs
        + diesel_ghgs
        + heat_exchanger_ghgs
        + heat_exchanger_installation_ghgs
        + hot_water_tank_ghgs
        + hot_water_tank_installation_ghgs
        + misc_ghgs
        + pv_ghgs
        + pv_installation_ghgs
        + pvt_ghgs
        + pvt_installation_ghgs
        + storage_ghgs
    )


def calculate_connections_ghgs(
    ghg_inputs: Dict[str, Any], households: pd.Series
) -> float:
    """
    Calculates ghgs of connecting households to the system

    Inputs:
        - ghg_inputs:
            GHG input information.
        - households:
            DataFrame of households from Energy_System().simulation(...)

    Outputs:
        GHGs

    """

    # Ensure that the correct type is being used.
    households_data_frame: pd.DataFrame = pd.DataFrame(households)

    # Compute the number of new households that were added to the system.
    new_connections: float = np.max(households_data_frame) - np.min(
        households_data_frame
    )

    # Calculate the associated ghgs.
    connection_ghgs: float = ghg_inputs[ImpactingComponent.HOUSEHOLDS.value][
        CONNECTION_GHGS
    ]
    connections_ghgs = float(connection_ghgs * new_connections)

    return connections_ghgs


def calculate_grid_extension_ghgs(
    ghg_inputs: Dict[str, Any], grid_extension_distance: float
) -> float:
    """
    Calculates ghgs of extending the grid network to a community

    Inputs:
        - ghg_inputs:
            GHG input information.
        - grid_extension_distance:
            Distance to the existing grid network

    Outputs:
        GHGs

    """

    return grid_extension_distance * float(
        ghg_inputs[ImpactingComponent.GRID.value][EXTENSION_GHGS]
    ) + float(ghg_inputs[ImpactingComponent.GRID.value][INFRASTRUCTURE_GHGS])


def _calculate_inverter_ghgs(  # pylint: disable=too-many-locals
    electric_yearly_load_statistics: pd.DataFrame,
    end_year: int,
    ghg_inputs: Dict[str, Any],
    location: Location,
    logger: Logger,
    scenario: Scenario,
    start_year: int,
) -> float:
    """
    Calculates ghgs of inverters based on load calculations

    Inputs:
        - electric_yearly_load_statistics:
            The yearly load statistics for the electric load.
        - end_year:
            End year of simulation period
        - ghg_inputs:
            GHG input information.
        - location:
            The location being considered.
        - logger:
            The :class:`logging.Logger` to use for the run.
        - scenario:
            The :class:`Scenario` currently being considered.
        - start_year:
            Start year of simulation period

    Outputs:
        GHGs

    """

    # Calcualte inverter replacement periods
    replacement_period = int(ghg_inputs[ImpactingComponent.INVERTER.value][LIFETIME])
    replacement_intervals = pd.DataFrame(
        np.arange(0, location.max_years, replacement_period)
    )
    replacement_intervals.columns = pd.Index([ColumnHeader.INSTALLATION_YEAR.value])

    # Check if inverter should be replaced in the specified time interval
    if not any(
        replacement_intervals[ColumnHeader.INSTALLATION_YEAR.value].isin(
            list(np.array(range(start_year, end_year)))
        )
    ):
        return float(0.0)

    # Initialise inverter sizing calculation
    max_power = []
    inverter_step = float(ghg_inputs[ImpactingComponent.INVERTER.value][SIZE_INCREMENT])
    inverter_size: List[float] = []
    for i in range(len(replacement_intervals)):
        # Calculate maximum power in interval years
        start = replacement_intervals[ColumnHeader.INSTALLATION_YEAR.value].iloc[i]
        end = start + replacement_period
        max_power_interval = (
            electric_yearly_load_statistics[ColumnHeader.MAXIMUM.value]
            .iloc[start:end]
            .max()
        )
        max_power.append(max_power_interval)

        # Calculate resulting inverter size
        inverter_size_interval: float = (
            np.ceil(0.001 * max_power_interval / inverter_step) * inverter_step
        )
        inverter_size.append(inverter_size_interval)

    inverter_size_data_frame: pd.DataFrame = pd.DataFrame(inverter_size)
    inverter_size_data_frame.columns = pd.Index([ColumnHeader.INVERTER_SIZE.value])
    inverter_info = pd.concat([replacement_intervals, inverter_size_data_frame], axis=1)

    # Calculate the associated ghgs
    inverter_info["Inverter ghgs (kgCO2/kW)"] = [
        ghg_inputs[ImpactingComponent.INVERTER.value][GHGS]
        * (1 - 0.01 * ghg_inputs[ImpactingComponent.INVERTER.value][GHG_DECREASE])
        ** inverter_info[ColumnHeader.INSTALLATION_YEAR.value].iloc[i]
        for i in range(len(inverter_info))
    ]

    # If a static inverter size has been used, use this, otherwise, use the dynamically
    # calculated values.
    if scenario.fixed_inverter_size and any(
        inverter_info[ColumnHeader.INVERTER_SIZE.value] > scenario.fixed_inverter_size
    ):
        logger.info(
            "The static inverter size specified, %s, was below that calculated "
            "within CLOVER. Calculated inverter sizes:\n%s",
            scenario.fixed_inverter_size,
            "\n".join(
                [
                    f"Size for year {year}: {size} kW"
                    for year, size in zip(
                        replacement_intervals[ColumnHeader.INSTALLATION_YEAR.value],
                        inverter_info[ColumnHeader.INVERTER_SIZE.value],
                    )
                ]
            ),
        )

    inverter_info[ColumnHeader.TOTAL_GHGS.value] = [
        (
            inverter_info[ColumnHeader.INVERTER_SIZE.value].iloc[i]
            if not scenario.fixed_inverter_size
            else scenario.fixed_inverter_size
        )
        * inverter_info["Inverter ghgs (kgCO2/kW)"].iloc[i]
        for i in range(len(inverter_info))
    ]
    inverter_ghgs: float = np.sum(
        inverter_info.iloc[
            inverter_info.index[
                inverter_info[ColumnHeader.INSTALLATION_YEAR.value].isin(
                    list(np.array(range(start_year, end_year)))
                )
            ]
        ][ColumnHeader.TOTAL_GHGS.value]
    ).round(2)

    return inverter_ghgs


def calculate_independent_ghgs(
    electric_yearly_load_statistics: pd.DataFrame,
    end_year: int,
    ghg_inputs: Dict[str, Any],
    location: Location,
    logger: Logger,
    scenario: Scenario,
    start_year: int,
) -> float:
    """
    Calculates ghgs of equipment which is independent of simulation periods

    Inputs:
        - electric_yearly_load_statistics:
            The electric yearly load statistics for the simulation period.
        - end_year:
            End year of simulation period
        - ghg_inputs:
            The GHG input informaiton.
        - location:
            The location being considered.
        - logger:
            The :class:`logging.Logger` to use for the run.
        - scenario:
            The :class:`Scenario` currently being considered.
        - start_year:
            Start year of simulation period

    Outputs:
        GHGs

    """

    inverter_ghgs = _calculate_inverter_ghgs(
        electric_yearly_load_statistics,
        end_year,
        ghg_inputs,
        location,
        logger,
        scenario,
        start_year,
    )
    total_ghgs = inverter_ghgs  # ... + other components as required

    return total_ghgs


def calculate_kerosene_ghgs(
    ghg_inputs: Dict[str, Any], kerosene_lamps_in_use_hourly: pd.Series
) -> float:
    """
    Calculates ghgs of kerosene usage.

    Inputs:
        - kerosene_lamps_in_use_hourly:
            Output from Energy_System().simulation(...)

    Outputs:
        GHGs

    """

    kerosene_ghgs = (
        kerosene_lamps_in_use_hourly
        * ghg_inputs[ImpactingComponent.KEROSENE.value][GHGS]
    )

    return float(np.sum(kerosene_ghgs))


def calculate_kerosene_ghgs_mitigated(
    ghg_inputs: Dict[str, Any], kerosene_lamps_mitigated_hourly: pd.Series
) -> float:
    """
    Calculates ghgs of kerosene usage that has been avoided by using the system.

    Inputs:
        - kerosene_lamps_mitigated_hourly:
            Output from Energy_System().simulation(...)

    Outputs:
        GHGs

    """

    kerosene_ghgs = (
        kerosene_lamps_mitigated_hourly
        * ghg_inputs[ImpactingComponent.KEROSENE.value][GHGS]
    )

    return float(np.sum(kerosene_ghgs))


def calculate_grid_ghgs(
    ghg_inputs: Dict[str, Any],
    grid_energy_hourly: pd.Series,
    location: Location,
    start_year: int = 0,
    end_year: int = 20,
) -> float:
    """
    Calculates ghgs of grid electricity used by the system

    Inputs:
        - ghg_inputs:
            The GHG inputs.
        - grid_energy_hourly:
            Output from Energy_System().simulation(...)
        - location:
            The location currently being considered.
        - start_year:
            Start year of simulation period
        - end_year:
            End year of simulation period

    Outputs:
        GHGs

    """
    # Initialise
    grid_ghgs_initial = ghg_inputs[ImpactingComponent.GRID.value][INITIAL_GHGS]
    grid_ghgs_final = ghg_inputs[ImpactingComponent.GRID.value][FINAL_GHGS]
    days = int(365 * (end_year - start_year))
    total_daily_energy = pd.DataFrame(
        hourly_profile_to_daily_sum(pd.DataFrame(grid_energy_hourly.dropna()))  # type: ignore
    )

    # Account for reduction in grid GHG intensity
    yearly_decrease = (grid_ghgs_initial - grid_ghgs_final) / location.max_years

    # Compute the daily decrease:
    #   daily_decrease = yearly_decrease / days
    grid_ghgs_start = grid_ghgs_initial - start_year * yearly_decrease
    grid_ghgs_end = grid_ghgs_initial - end_year * yearly_decrease
    daily_emissions_intensity = pd.DataFrame(
        np.linspace(grid_ghgs_start, grid_ghgs_end, days)
    )

    # Calculate daily emissions
    daily_emissions = pd.DataFrame(
        total_daily_energy.values * daily_emissions_intensity.values
    )

    return float(np.sum(daily_emissions, axis=0))  # type: ignore


def calculate_diesel_fuel_ghgs(
    diesel_fuel_usage_hourly: pd.Series, ghg_inputs: Dict[str, Any]
) -> float:
    """
    Calculates ghgs of diesel fuel used by the system

    Inputs:
        - diesel_fuel_usage_hourly:
            Output from Energy_System().simulation(...)
        - ghg_inputs:
            GHG input information.

    Outputs:
        GHGs

    """

    diesel_fuel_ghgs = ghg_inputs[ImpactingComponent.DIESEL_FUEL.value][GHGS]
    return float(np.sum(diesel_fuel_usage_hourly) * diesel_fuel_ghgs)  # type: ignore


def calculate_om_ghgs(
    capacity: float,
    ghg_inputs: Dict[str, Any],
    system_component: str,
    start_year: int = 0,
    end_year: int = 20,
) -> float:
    """
    Calculates the O&M GHGs of a component.

    Inputs:
        - capacity:
            The capacity of the relevant component.
        - ghg_inputs:
            GHG input data.
        - system_component:
            The component of the system currently being considered, as a `str`
            representing the component rather than the component.
        - start_year:
            The start year for the simulation.
        - end_year:
            The end year for the simulation.

    Outputs:
        GHGs

    """

    return (
        capacity
        * float(ghg_inputs[system_component][OM_GHGS])
        * (end_year - start_year)
    )


#   Total O&M for entire system
def calculate_total_om(  # pylint: disable=too-many-locals
    buffer_tanks: int,
    clean_water_tanks: int,
    converters: Optional[Dict[str, int]],
    diesel_size: float,
    ghg_inputs: Dict[str, Any],
    heat_exchangers: int,
    hot_water_tanks: int,
    logger: Logger,
    pv_array_size: float,
    pvt_array_size: float,
    storage_size: float,
    start_year: int = 0,
    end_year: int = 20,
) -> float:
    """
    Calculates total O&M ghgs over the simulation period

    Inputs:
        - buffer_tanks:
            Capacity of buffer tanks installed.
        - clean_water_tanks:
            Capacity of clean-water tanks installed.
        - converters:
            A mapping between converter names and the size of each that was added to the
            system this iteration.
        - diesel_size:
            Capacity of diesel generator installed.
        - ghg_inputs:
            The GHG input information.
        - heat_exchangers:
            Capacity of heat exchangers installed.
        - hot_water_tanks:
            Capacity of hot-water tanks installed.
        - logger:
            The :class:`logging.Logger` to use for the run.
        - pv_array_size:
            Capacity of PV installed.
        - pvt_array_size:
            Capacity of PV-T installed.
        - storage_size:
            Capacity of battery storage installed.
        - start_year:
            Start year of simulation period.
        - end_year:
            End year of simulation period.

    Outputs:
        GHGs

    """

    if ImpactingComponent.BUFFER_TANK.value not in ghg_inputs and buffer_tanks > 0:
        logger.error(
            "%sNo buffer tank GHG input information provided.%s",
            BColours.fail,
            BColours.endc,
        )
        raise InputFileError(
            "tank inputs",
            "No buffer tank ghg input information provided and a non-zero number of "
            "buffer tanks are being considered.",
        )
    buffer_tank_om_ghgs: float = 0
    if buffer_tanks > 0:
        buffer_tank_om_ghgs = calculate_om_ghgs(
            buffer_tanks,
            ghg_inputs,
            ImpactingComponent.BUFFER_TANK.value,
            start_year,
            end_year,
        )

    if (
        ImpactingComponent.CLEAN_WATER_TANK.value not in ghg_inputs
        and clean_water_tanks > 0
    ):
        logger.error(
            "%sNo clean-water tank GHG input information provided.%s",
            BColours.fail,
            BColours.endc,
        )
        raise InputFileError(
            "tank inputs",
            "No clean-water tank ghg input information provided and a non-zero number "
            "of clean-water tanks are being considered.",
        )
    clean_water_tank_om_ghgs: float = 0
    if clean_water_tanks > 0:
        clean_water_tank_om_ghgs = calculate_om_ghgs(
            clean_water_tanks,
            ghg_inputs,
            ImpactingComponent.CLEAN_WATER_TANK.value,
            start_year,
            end_year,
        )

    converter_om_ghgs: float = 0
    if converters is not None:
        converter_om_ghgs = sum(
            calculate_om_ghgs(
                size,
                ghg_inputs,
                GHG_IMPACT.format(
                    type=ImpactingComponent.CONVERTER.value, name=converter
                ),
                start_year,
                end_year,
            )
            for converter, size in converters.items()
        )
    else:
        logger.debug("No converters installed so no converter OM GHGs to calcualte.")

    diesel_om_ghgs = calculate_om_ghgs(
        diesel_size, ghg_inputs, ImpactingComponent.DIESEL.value, start_year, end_year
    )

    general_om_ghgs = calculate_om_ghgs(
        1, ghg_inputs, ImpactingComponent.GENERAL.value, start_year, end_year
    )

    if (
        ImpactingComponent.HEAT_EXCHANGER.value not in ghg_inputs
        and heat_exchangers > 0
    ):
        logger.error(
            "%sNo heat-exchanger GHG input information provided.%s",
            BColours.fail,
            BColours.endc,
        )
        raise InputFileError(
            "heat exchanger inputs",
            "No heat-exchanger ghg input information provided and a non-zero number of "
            "heat exchangers are being considered.",
        )
    heat_exchanger_om_ghgs: float = 0
    if heat_exchangers > 0:
        heat_exchanger_om_ghgs = calculate_om_ghgs(
            heat_exchangers,
            ghg_inputs,
            ImpactingComponent.HEAT_EXCHANGER.value,
            start_year,
            end_year,
        )

    if (
        ImpactingComponent.HOT_WATER_TANK.value not in ghg_inputs
        and hot_water_tanks > 0
    ):
        logger.error(
            "%sNo hot-water tank GHG input information provided.%s",
            BColours.fail,
            BColours.endc,
        )
        raise InputFileError(
            "tank inputs",
            "No hot-water tank ghg input information provided and a non-zero number of "
            "hot-water tanks are being considered.",
        )
    hot_water_tank_om_ghgs: float = 0
    if hot_water_tanks > 0:
        hot_water_tank_om_ghgs = calculate_om_ghgs(
            hot_water_tanks,
            ghg_inputs,
            ImpactingComponent.HOT_WATER_TANK.value,
            start_year,
            end_year,
        )

    pv_om_ghgs = calculate_om_ghgs(
        pv_array_size, ghg_inputs, ImpactingComponent.PV.value, start_year, end_year
    )

    if ImpactingComponent.PV_T.value not in ghg_inputs and pvt_array_size > 0:
        logger.error(
            "%sNo PV-T GHG input information provided.%s", BColours.fail, BColours.endc
        )
        raise InputFileError(
            "solar generation inputs",
            "No PV-T ghg input information provided and a non-zero number of PV-T"
            "panels are being considered.",
        )
    pvt_om_ghgs: float = 0
    if pvt_array_size > 0:
        pvt_om_ghgs = calculate_om_ghgs(
            pvt_array_size,
            ghg_inputs,
            ImpactingComponent.PV_T.value,
            start_year,
            end_year,
        )
    storage_om_ghgs = calculate_om_ghgs(
        storage_size, ghg_inputs, ImpactingComponent.STORAGE.value, start_year, end_year
    )

    return (
        buffer_tank_om_ghgs
        + clean_water_tank_om_ghgs
        + converter_om_ghgs
        + diesel_om_ghgs
        + general_om_ghgs
        + heat_exchanger_om_ghgs
        + hot_water_tank_om_ghgs
        + pv_om_ghgs
        + pvt_om_ghgs
        + storage_om_ghgs
    )



#!/usr/bin/python3
########################################################################################
# __init__.py - Python internals module, no functionality here.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
# Date created: 05/07/2021
# License: Open source
########################################################################################
"""
mains_supply.__init__.py - Init module for the mains-supply component.

The `__init__` module ensures that all of the packages are correctly exposed so that
they can be imported when CLOVER is installed as a package.

"""

from .__utils__ import get_intermittent_supply_status
from .grid import get_lifetime_grid_status, load_grid_profile
from .water_source import get_lifetime_water_source_status



#!/usr/bin/python3
########################################################################################
# storage.py - Storage module.                                                         #
#                                                                                      #
# Author: Ben Winchester                                                               #
# Copyright: Ben Winchester, 2020                                                      #
# Date created: 16/07/2021                                                             #
# License: Open source                                                                 #
# Most recent update: 16/07/2021                                                       #
########################################################################################
"""
storage.py - The storage module for CLOVER.

CLOVER considers several storage media for various forms of energy. The various
calculations associated with these storage media are carried out in this module.

"""

from logging import Logger
from typing import Dict, Optional, Tuple

import pandas as pd
import numpy as np

from ..__utils__ import (
    BColours,
    CleanWaterMode,
    ColumnHeader,
    DistributionNetwork,
    InputFileError,
    InternalError,
    Location,
    RenewableEnergySource,
    Scenario,
)
from .__utils__ import Minigrid
from ..conversion.conversion import WaterSource
from ..generation.solar import solar_degradation

__all__ = (
    "battery_iteration_step",
    "cw_tank_iteration_step",
    "get_electric_battery_storage_profile",
    "get_water_storage_profile",
)


def battery_iteration_step(
    battery_storage_profile: pd.DataFrame,
    hourly_battery_storage: Dict[int, float],
    initial_battery_storage: float,
    logger: Logger,
    maximum_battery_storage: float,
    minigrid: Minigrid,
    minimum_battery_storage: float,
    *,
    time_index: int,
) -> Tuple[float, float, float]:
    """
    Carries out an iteration calculation for the battery.

    Inputs:
        - battery_storage_profile:
            The battery storage profile, as a :class:`pandas.DataFrame`, giving the net
            flow into and out of the battery due to renewable electricity generation.
        - hourly_battery_storage:
            The mapping between time and computed battery storage.
        - initial_battery_storage:
            The initial amount of energy stored in the batteries.
        - logger:
            The :class:`logging.Logger` to use for the run.
        - maximum_battery_storage:
            The maximum amount of energy that can be stored in the batteries.
        - minigrid:
            The :class:`Minigrid` representing the system being considered.
        - minimum_battery_storage:
            The minimum amount of energy that can be stored in the batteries.
        - time_index:
            The current time (hour) being considered.

    Outputs:
        - battery_energy_flow:
            The net flow into or out of the battery.
        - excess_energy:
            The energy surplus generated which could not be stored in the batteries.
        - new_hourly_battery_storage;
            The computed level of energy stored in the batteries at this time step.

    """

    if minigrid.battery is None:
        logger.error(
            "%sNo battery was defined on the minigrid despite the iteration "
            "calculation being called to compute the energy stored within the "
            "batteries. Either define a valid battery for the energy system, or adjust "
            "the scenario to no longer consider battery inputs.%s",
            BColours.fail,
            BColours.endc,
        )
        raise InputFileError(
            "energy system inputs",
            "Battery undefined despite an itteration step being called.",
        )

    battery_energy_flow = battery_storage_profile.iloc[time_index, 0]
    if time_index == 0:
        new_hourly_battery_storage = initial_battery_storage + battery_energy_flow
    else:
        # Battery charging
        if battery_energy_flow >= 0.0:
            new_hourly_battery_storage = hourly_battery_storage[time_index - 1] * (
                1.0 - minigrid.battery.leakage
            ) + minigrid.battery.conversion_in * min(
                battery_energy_flow,
                minigrid.battery.charge_rate
                * (maximum_battery_storage - minimum_battery_storage),
            )
        # Battery discharging
        else:
            new_hourly_battery_storage = hourly_battery_storage[time_index - 1] * (
                1.0 - minigrid.battery.leakage
            ) + (1.0 / minigrid.battery.conversion_out) * max(
                battery_energy_flow,
                (-1.0)
                * minigrid.battery.discharge_rate
                * (maximum_battery_storage - minimum_battery_storage),
            )

    excess_energy = max(new_hourly_battery_storage - maximum_battery_storage, 0.0)

    return battery_energy_flow, excess_energy, new_hourly_battery_storage


def cw_tank_iteration_step(  # pylint: disable=too-many-locals
    backup_desalinator_water_supplied: Dict[int, float],
    clean_water_power_consumed_mapping: Dict[int, float],
    clean_water_demand_met_by_excess_energy: Dict[int, float],
    clean_water_supplied_by_excess_energy: Dict[int, float],
    conventional_cw_source_profiles: Optional[Dict[WaterSource, pd.DataFrame]],
    conventional_water_supplied: Dict[int, float],
    energy_per_desalinated_litre: float,
    excess_energy: float,
    excess_energy_used_desalinating: Dict[int, float],
    hourly_cw_tank_storage: Dict[int, float],
    initial_cw_tank_storage: float,
    logger: Logger,
    maximum_battery_storage: float,
    maximum_cw_tank_storage: float,
    maximum_water_throughput: float,
    minigrid: Minigrid,
    minimum_cw_tank_storage: float,
    new_hourly_battery_storage: float,
    scenario: Scenario,
    storage_water_supplied: Dict[int, float],
    tank_storage_profile: pd.DataFrame,
    *,
    time_index: int,
) -> float:
    """
    Caries out an iteration calculation for the clean-water tanks.

    Inputs:
        - backup_desalinator_water_supplied:
            The water supplied by the backup (electric) desalination.
        - clean_water_power_consumed_mapping:
            The power consumed in providing clean water.
        - clean_water_demand_met_by_excess_energy:
            The clean-water demand that was met through excess energy from the renewable
            system.
        - clean_water_supplied_by_excess_energy:
            The clean water that was supplied by the excess energy from the renewable
            system.
        - conventioanl_cw_source_profiles:
            A mapping between :class:`WaterSource` instances, corresponding to
            conventional sources of drinking water within the system, and their
            associated maximum output throughout the duration of the simulation.
        - conventional_water_supplied:
            A mapping between time index and the amount of clean water supplied through
            conventional sources available to the system.
        - energy_per_desalinated_litre:
            The electrical energy required to desalinate a single litre.
        - excess_energy:
            The excess electrical energy from the renewable system.
        - excess_energy_used_desalinating:
            The amount of excess electrical energy that was used desalinating.
        - hourly_cw_tank_storage:
            A mapping between time index and the amount of clean water stored in the
            system.
        - initial_cw_tank_storage:
            The initial level of the clean water tanks.
        - maximum_battery_storage:
            The maximum amount of energy that can be stored in the batteries.
        - maximum_cw_tank_storage:
            The maximum storage of the clean-water tanks.
        - maximum_water_throughput:
            The maximum amount of water that can be desalinated electrically.
        - minigrid:
            The :class:`Minigrid` being used for the run.
        - minimum_cw_tank_storage:
            The minimum amount of water that must be held in the clean-water tanks.
        - new_hourly_battery_storage:
            The level of electricity stored in the batteries at the time step being
            considered.
        - scenario:
            The :class:`Scenario` for the run being carried out.
        - storage_water_supplied:
            The amount of clean water, in litres, that was supplied by the clean-water
            storage tanks.
        - time_index:
            The current index being considered.

    Outputs:
        - excess_energy:
            The excess electrical energy, generated by the renewables, after what can be
            used for desalination has been used for electrical desalination.

    """

    if scenario.desalination_scenario is not None:
        tank_water_flow = tank_storage_profile.iloc[time_index, 0]

        # Raise an error if there is no clean-water tank defined.
        if minigrid.clean_water_tank is not None:
            logger.error(
                "%sNo clean-water tank defined despite desalination being carried out"
                ".%s",
                BColours.fail,
                BColours.endc,
            )
            raise InputFileError(
                "minigrid inputs",
                "No clean-water tank defined despite desalination being modelled.",
            )

        # Compute the new tank level based on the previous level and the flow.
        if time_index == 0:
            current_net_water_flow = initial_cw_tank_storage + tank_water_flow
        else:
            current_net_water_flow = (
                hourly_cw_tank_storage[time_index - 1]
                * (1.0 - minigrid.clean_water_tank.leakage)  # type: ignore
                + tank_water_flow
            )

        # Use the excess energy to desalinate if there is space.
        if (
            excess_energy > 0
            and scenario.desalination_scenario is not None
            and scenario.desalination_scenario.clean_water_scenario.mode
            == CleanWaterMode.BACKUP
        ):
            # Compute the maximum amount of water that can be desalinated.
            maximum_desalinated_water = min(
                excess_energy / energy_per_desalinated_litre,
                maximum_water_throughput,
            )

            # Add this to the tank and fulfil the demand if relevant.
            current_hourly_cw_tank_storage = (
                current_net_water_flow + maximum_desalinated_water
            )

            # Compute the amount of water that was actually desalinated.
            desalinated_water = min(
                maximum_desalinated_water,
                maximum_cw_tank_storage - current_net_water_flow,
            )

            # Compute the remaining excess energy and the energy used in
            # desalination.
            energy_consumed = energy_per_desalinated_litre * desalinated_water
            new_hourly_battery_storage -= energy_consumed

            # Ensure that the excess energy is normalised correctly.
            excess_energy = max(
                new_hourly_battery_storage - maximum_battery_storage, 0.0
            )

            # Store this as water and electricity supplied using excess power.
            excess_energy_used_desalinating[time_index] = energy_consumed
            clean_water_demand_met_by_excess_energy[time_index] = max(
                0, -current_net_water_flow
            )
            clean_water_supplied_by_excess_energy[time_index] = desalinated_water
        else:
            excess_energy_used_desalinating[time_index] = 0
            clean_water_demand_met_by_excess_energy[time_index] = 0
            clean_water_supplied_by_excess_energy[time_index] = 0
            current_hourly_cw_tank_storage = current_net_water_flow

        # If there is still unmet water demand, then carry out desalination and
        # pumping to fulfil the demand.
        current_unmet_water_demand: float = -current_hourly_cw_tank_storage
        if (
            current_unmet_water_demand > 0
            and scenario.desalination_scenario is not None
            and scenario.desalination_scenario.clean_water_scenario.mode
            == CleanWaterMode.PRIORITISE
        ):
            # Compute the electricity consumed meeting this demand.
            energy_consumed = energy_per_desalinated_litre * current_unmet_water_demand

            # Withdraw this energy from the batteries.
            new_hourly_battery_storage -= (
                1.0 / minigrid.battery.conversion_out  # type: ignore
            ) * energy_consumed

            # Ensure that the excess energy is normalised correctly.
            excess_energy = max(
                new_hourly_battery_storage - maximum_battery_storage, 0.0
            )

            # Store this as water and electricity supplied by backup.
            clean_water_power_consumed_mapping[time_index] += energy_consumed
            backup_desalinator_water_supplied[time_index] = current_unmet_water_demand
        else:
            clean_water_power_consumed_mapping[time_index] = 0
            backup_desalinator_water_supplied[time_index] = 0

        # Any remaining unmet water demand should be met using conventional clean-water
        # sources if available.
        if current_unmet_water_demand > 0:
            # Compute the clean water supplied using convnetional sources.
            conventional_cw_available: float = 0
            if conventional_cw_source_profiles is not None:
                conventional_cw_available = float(
                    sum(  # type: ignore [arg-type]
                        entry.iloc[time_index]
                        for entry in conventional_cw_source_profiles.values()
                    )
                )
            conventional_cw_supplied = min(
                conventional_cw_available, current_unmet_water_demand
            )
            current_unmet_water_demand -= conventional_cw_supplied

            # Store this as water supplied through conventional means.
            conventional_water_supplied[time_index] = conventional_cw_supplied
        else:
            conventional_water_supplied[time_index] = 0

        current_hourly_cw_tank_storage = min(
            current_hourly_cw_tank_storage,
            maximum_cw_tank_storage,
        )
        current_hourly_cw_tank_storage = max(
            current_hourly_cw_tank_storage,
            minimum_cw_tank_storage,
        )

        hourly_cw_tank_storage[time_index] = current_hourly_cw_tank_storage

        if time_index == 0:
            storage_water_supplied[time_index] = 0.0 - tank_water_flow
        else:
            storage_water_supplied[time_index] = max(
                hourly_cw_tank_storage[time_index - 1]
                * (1.0 - minigrid.clean_water_tank.leakage)  # type: ignore
                - hourly_cw_tank_storage[time_index],
                0.0,
            )

    return excess_energy


def get_electric_battery_storage_profile(  # pylint: disable=too-many-locals, too-many-statements
    *,
    grid_profile: pd.Series,
    kerosene_usage: pd.Series,
    location: Location,
    logger: Logger,
    minigrid: Minigrid,
    processed_total_electric_load: pd.DataFrame,
    renewables_power_produced: Dict[RenewableEnergySource, pd.DataFrame],
    scenario: Scenario,
    clean_water_pvt_size: int = 0,
    end_hour: int = 4,
    hot_water_pvt_size: int = 0,
    pv_size: float = 10,
    start_hour: int = 0,
) -> Tuple[
    pd.DataFrame,
    pd.DataFrame,
    pd.Series,
    pd.DataFrame,
    pd.DataFrame,
    Dict[RenewableEnergySource, pd.DataFrame],
    pd.DataFrame,
]:
    """
    Gets the storage profile (energy in/out the battery) and other system energies.

    Inputs:
        - grid_profile:
            The relevant grid profile, based on the scenario, for the simulation.
        - kerosene_usage:
            The kerosene usage.
        - logger:
            The logger to use for the run.
        - minigrid:
            The energy system being modelled.
        - processed_total_electric_load:
            The total electric load for the system.
        - renewables_power_produced:
            The total electric power produced, per renewable type, as a mapping between
            :class:`SolarPanelType` and :class:`pandas.DataFrame` instances, with units
            of technology size.
        - scenario:
            The scenatio being considered.
        - clean_water_pvt_size:
            Amount of PV-T in units of PV-T associated with the clean-water system.
        - end_year:
            End year of this simulation period.
        - hot_water_pvt_size:
            Amount of PV-T in units of PV-T associated with the hot-water system.
        - pv_size:
            Amount of PV in units of PV.
        - start_year:
            Start year of this simulation period.

    Outputs:
        - battery_storage_profile:
            Amount of energy (kWh) into (+ve) and out of (-ve) the battery.
        - grid_energy:
            Amount of energy (kWh) supplied by the grid.
        - kerosene_usage:
            Number of kerosene lamps in use (if no power available).
        - load_energy:
            Amount of energy (kWh) required to satisfy the loads.
        - pvt_energy:
            Amount of energy (kWh) provided by PV to the system.
        - pvt_energy:
            Amount of electric energy (kWh) provided by PV-T to the system.
        - renewables_energy:
            Amount of energy (kWh) provided by renewables to the system.
        - renewables_energy_map:
            A mapping between :class:`RenewableEnergySource` and the associated
            electrical energy produced.
        - renewables_energy_used_directly:
            Amount of energy (kWh) from renewables used directly to satisfy load (kWh).

    """

    # Initialise power generation, including degradation of PV
    try:
        pv_power_produced = renewables_power_produced[RenewableEnergySource.PV]
    except KeyError:
        logger.critical(
            "%sCould not determine PV power produced from renewables production.%s",
            BColours.fail,
            BColours.endc,
        )
        raise InternalError(
            "No PV power in renewables_power_produced mapping, fatal."
        ) from None
    pv_generation_array = pv_power_produced * pv_size
    solar_degradation_array = solar_degradation(
        minigrid.pv_panel.lifetime, location.max_years
    ).iloc[start_hour:end_hour, 0]
    pv_generation = pd.DataFrame(
        np.asarray(pv_generation_array.iloc[start_hour:end_hour])
        * np.asarray(solar_degradation_array)
    )

    # Initialise PV-T power generation, including degradation of PV
    if minigrid.pvt_panel is not None:
        # Determine the PV-T degredation.
        pvt_degradation_array = solar_degradation(  # type: ignore
            minigrid.pvt_panel.lifetime, location.max_years
        )[0 : (end_hour - start_hour)]

        if (
            RenewableEnergySource.CLEAN_WATER_PVT not in renewables_power_produced
            and RenewableEnergySource.HOT_WATER_PVT not in renewables_power_produced
        ):
            logger.error(
                "%sA PV-T panel was defined on the system but no clean-water PV-T or "
                "hot-water PV-T electricity was generated.%s",
                BColours.fail,
                BColours.endc,
            )
            raise InternalError(
                "No PV-T electric power produced despite a PV-T panel being defined "
                "for the system.."
            )

        # Compute the clean-water PV-T electricity generated.
        if RenewableEnergySource.CLEAN_WATER_PVT in renewables_power_produced:
            try:
                clean_water_pvt_electric_power_produced = renewables_power_produced[
                    RenewableEnergySource.CLEAN_WATER_PVT
                ]
            except KeyError:
                logger.error(
                    "%sCould not determine clean-water PV-T power produced from "
                    "renewables production despite a PV-T panel being defined on the "
                    "system.%s",
                    BColours.fail,
                    BColours.endc,
                )
                raise InternalError(
                    "No PV-T power in renewables_power_produced mapping despite a PV-T "
                    "panel being specified."
                ) from None
            clean_water_pvt_electric_generation_array = (
                clean_water_pvt_electric_power_produced * clean_water_pvt_size
            )
            clean_water_pvt_electric_generation: pd.DataFrame = pd.DataFrame(
                np.asarray(clean_water_pvt_electric_generation_array)
                * np.asarray(pvt_degradation_array)
            )
        else:
            clean_water_pvt_electric_generation = pd.DataFrame(
                [0] * (end_hour - start_hour)
            )

        # Compute the clean-water source.
        if RenewableEnergySource.HOT_WATER_PVT in renewables_power_produced:
            try:
                hot_water_pvt_electric_power_produced = renewables_power_produced[
                    RenewableEnergySource.HOT_WATER_PVT
                ]
            except KeyError:
                logger.error(
                    "%sCould not determine PV-T power produced from renewables "
                    "production despite a PV-T panel being defined on the system.%s",
                    BColours.fail,
                    BColours.endc,
                )
                raise InternalError(
                    "No PV-T power in renewables_power_produced mapping despite a PV-T "
                    "panel being specified."
                ) from None
            hot_water_pvt_electric_generation_array = (
                hot_water_pvt_electric_power_produced * hot_water_pvt_size
            )
            hot_water_pvt_electric_generation: pd.DataFrame = pd.DataFrame(
                np.asarray(hot_water_pvt_electric_generation_array)
                * np.asarray(pvt_degradation_array)
            )
        else:
            hot_water_pvt_electric_generation = pd.DataFrame(
                [0] * (end_hour - start_hour)
            )

    else:
        clean_water_pvt_electric_generation = pd.DataFrame(
            [0] * (end_hour - start_hour)
        )
        hot_water_pvt_electric_generation = pd.DataFrame([0] * (end_hour - start_hour))

    # Consider power distribution network
    if scenario.distribution_network == DistributionNetwork.DC:
        pv_generation = pv_generation.mul(  # type: ignore
            minigrid.dc_to_dc_conversion_efficiency
        )
        transmission_efficiency = minigrid.dc_transmission_efficiency
        # grid_conversion_eff = minigrid.ac_to_dc_conversion

    else:
        pv_generation = pv_generation.mul(  # type: ignore
            minigrid.dc_to_ac_conversion_efficiency
        )
        transmission_efficiency = minigrid.ac_transmission_efficiency
        # grid_conversion_efficiency = minigrid.ac_to_ac_conversion

    if transmission_efficiency is None:
        logger.error(
            "%sNo valid transmission efficiency was determined based on the energy "
            "system inputs. Check this before continuing.%s",
            BColours.fail,
            BColours.endc,
        )
        raise InputFileError(
            "energy system inputs",
            "No valid transmission efficiency was determined based on the energy "
            "system inputs. Check this before continuing.",
        )

    # Consider transmission efficiency
    load_energy: pd.DataFrame = (
        processed_total_electric_load / transmission_efficiency  # type: ignore
    )
    pv_energy = pv_generation * transmission_efficiency

    if clean_water_pvt_electric_generation is not None:
        pvt_cw_electric_energy: pd.DataFrame = (
            clean_water_pvt_electric_generation * transmission_efficiency
        )
    else:
        pvt_cw_electric_energy = pd.DataFrame([0] * pv_energy.size)

    if hot_water_pvt_electric_generation is not None:
        pvt_hw_electric_energy: pd.DataFrame = (
            hot_water_pvt_electric_generation * transmission_efficiency
        )
    else:
        pvt_hw_electric_energy = pd.DataFrame([0] * pv_energy.size)

    # Combine energy from all renewables sources
    renewables_energy_map: Dict[RenewableEnergySource, pd.DataFrame] = {
        RenewableEnergySource.PV: pv_energy,
        RenewableEnergySource.CLEAN_WATER_PVT: pvt_cw_electric_energy,
        RenewableEnergySource.HOT_WATER_PVT: pvt_hw_electric_energy,
        # RenewableGenerationSource.WIND: wind_energy, etc.
    }

    # Add more renewable sources here as required
    renewables_energy: pd.DataFrame = pd.DataFrame(
        sum(renewables_energy_map.values())  # type: ignore
    )

    # Check for self-generation prioritisation
    if scenario.prioritise_self_generation:
        # Take energy from PV first
        remaining_profile = pd.DataFrame(renewables_energy.values - load_energy.values)
        renewables_energy_used_directly: pd.DataFrame = pd.DataFrame(
            (remaining_profile > 0) * load_energy.values
            + (remaining_profile < 0) * renewables_energy.values
        )

        # Then take energy from grid if available
        if scenario.grid:
            grid_energy: pd.DataFrame = pd.DataFrame(
                ((remaining_profile < 0) * remaining_profile).iloc[:, 0]  # type: ignore
                * -1.0
                * grid_profile.values
            )
        else:
            grid_energy = pd.DataFrame([0] * (end_hour - start_hour))
        battery_storage_profile: pd.DataFrame = pd.DataFrame(
            remaining_profile.values + grid_energy.values
        )

    else:
        # Take energy from grid first if available
        if scenario.grid:
            grid_energy = pd.DataFrame(grid_profile.mul(load_energy[0]))  # type: ignore
        else:
            grid_energy = pd.DataFrame([0] * (end_hour - start_hour))
        # as needed for load
        remaining_profile = (grid_energy[0] <= 0).mul(load_energy[0])  # type: ignore
        logger.debug(
            "Remainig profile: %s kWh",
            round(float(np.sum(remaining_profile)), 2),  # type: ignore [arg-type]
        )

        # Then take energy from PV if generated
        logger.debug(
            "Renewables profile: %s kWh",
            f"{round(float(np.sum(renewables_energy)), 2)}",  # type: ignore [arg-type]
        )
        battery_storage_profile = pd.DataFrame(
            renewables_energy[0].values - remaining_profile.values  # type: ignore
        )
        logger.debug(
            "Storage profile: %s kWh",
            f"{round(float(np.sum(battery_storage_profile)), 2)}",  # type: ignore [arg-type]
        )

        renewables_energy_used_directly = pd.DataFrame(
            ((renewables_energy[0] > 0) * (remaining_profile > 0))  # type: ignore [call-overload]
            * pd.concat(  # type: ignore [call-overload]
                [renewables_energy[0], remaining_profile], axis=1  # type: ignore [call-overload]
            ).min(axis=1)
        )

        logger.debug(
            "Grid energy: %s kWh",
            f"{round(float(np.sum(grid_energy)), 2)}",  # type: ignore [arg-type]
        )
        renewables_direct_rounded: float = round(
            float(np.sum(renewables_energy_used_directly)), 2  # type: ignore [arg-type]
        )
        logger.debug(
            "Renewables direct: %s kWh",
            round(float(np.sum(renewables_energy_used_directly)), 2),  # type: ignore [arg-type]
        )
        logger.debug("Renewables direct: %s kWh", renewables_direct_rounded)

    battery_storage_profile.columns = pd.Index([ColumnHeader.STORAGE_PROFILE.value])
    grid_energy.columns = pd.Index([ColumnHeader.GRID_ENERGY.value])
    load_energy.columns = pd.Index([ColumnHeader.LOAD_ENERGY.value])
    renewables_energy.columns = pd.Index(
        [ColumnHeader.RENEWABLE_ELECTRICITY_SUPPLIED.value]
    )
    renewables_energy_map[RenewableEnergySource.PV].columns = pd.Index(
        [ColumnHeader.PV_ELECTRICITY_SUPPLIED.value]
    )
    renewables_energy_map[RenewableEnergySource.CLEAN_WATER_PVT].columns = pd.Index(
        [ColumnHeader.CW_PVT_ELECTRICITY_SUPPLIED.value]
    )
    renewables_energy_map[RenewableEnergySource.HOT_WATER_PVT].columns = pd.Index(
        [ColumnHeader.HW_PVT_ELECTRICITY_SUPPLIED.value]
    )
    renewables_energy_used_directly.columns = pd.Index(
        [ColumnHeader.RENEWABLE_ELECTRICITY_USED_DIRECTLY.value]
    )

    return (
        battery_storage_profile,
        grid_energy,
        kerosene_usage,
        load_energy,
        renewables_energy,
        renewables_energy_map,
        renewables_energy_used_directly,
    )


def get_water_storage_profile(
    processed_total_cw_load: pd.DataFrame,
    renewable_cw_produced: pd.DataFrame,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Gets the storage profile for the clean-water system.

    Inputs:
        - minigrid:
            The minigrid being modelled.
        - processed_total_cw_load:
            The total clean-water load placed on the system.
        - renewable_cw_produced:
            The total clean water produced directly from renewables, i.e., solar-based
            or solar-thermal-based desalination technologies.
        - scenario:
            The scenario being considered.

    Outputs:
        - power_consumed:
            The electric power consumed in providing the water demand.
        - renewable_cw_used_directly:
            The renewable clean water which was directly consumed.
        - tank_storage_profile:
            The amount of water (litres) into (+ve) and out of (-ve) the clean-water
            tanks.

    """

    # Clean water is either produced directly or drawn from the storage tanks.
    remaining_profile = pd.DataFrame(
        renewable_cw_produced.values - processed_total_cw_load.values
    )
    renewable_cw_used_directly: pd.DataFrame = pd.DataFrame(
        (remaining_profile > 0) * processed_total_cw_load.values
        + (remaining_profile < 0) * renewable_cw_produced.values
    )

    tank_storage_profile: pd.DataFrame = pd.DataFrame(remaining_profile.values)

    electric_power_consumed: pd.DataFrame = 0.001 * pd.DataFrame(  # type: ignore
        [0] * processed_total_cw_load.size
    )

    return (
        electric_power_consumed,
        renewable_cw_used_directly,
        tank_storage_profile,
    )



#!/usr/bin/python3
########################################################################################
# __init__.py - Python internals module, no functionality here.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
# Date created: 05/07/2021
# License: Open source
########################################################################################
"""
load.__init__.py - Init module for the load component.

The `__init__` module ensures that all of the packages are correctly exposed so that
they can be imported when CLOVER is installed as a package.

"""

from .load import (
    compute_total_hourly_load,
    compute_processed_load_profile,
    DEFAULT_KEROSENE_DEVICE,
    Device,
    LOAD_LOGGER_NAME,
    ResourceType,
    population_hourly,
    process_device_hourly_power,
    process_device_hourly_usage,
    process_device_ownership,
    process_device_utilisation,
    process_load_profiles,
)



#!/usr/bin/python3
########################################################################################
# __init__.py - Python internals module, no functionality here.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
# Date created: 05/07/2021
# License: Open source
########################################################################################
"""
conversion.__init__.py - Init module for the conversion component.

The `__init__` module ensures that all of the packages are correctly exposed so that
they can be imported when CLOVER is installed as a package.

"""

from .conversion import (
    Converter,
    MultiInputConverter,
    ThermalDesalinationPlant,
    WaterSource,
)



#!/usr/bin/python3
########################################################################################
# water_source.py - Water source profile generation module.                            #
#                                                                                      #
# Author: Phil Sandwell, Ben Winchester                                                #
# Copyright: Phil Sandwell, 2018                                                       #
# License: Open source                                                                 #
# Most recent update: 01/11/2021                                                       #
#                                                                                      #
# For more information, please email:                                                  #
#     philip.sandwell@gmail.com                                                        #
########################################################################################
"""
water_source.py - The water-source generation module for CLOVER.

This module generates availability profiles for conventional water sources for CLOVER.

"""

import os

from logging import Logger
from typing import Dict

import pandas as pd  # pylint: disable=import-error

from tqdm import tqdm

from ..__utils__ import BColours, Location, monthly_times_to_daily_times, SKIPPING
from ..conversion.conversion import WaterSource

__all__ = ("get_lifetime_water_source_status",)


def _process_water_source_availability(
    water_source: WaterSource,
    *,
    availability: pd.DataFrame,
    generated_water_source_availability_directory: str,
    keyword: str,
    location: Location,
    logger: Logger,
    regenerate: bool,
) -> pd.DataFrame:
    """
    Process water-source data files.

    Processes the water-source data files, taking the availability probabilities and
    converting them to a :class:`pd.DataFrame` giving the availability of the water
    source for every hour of the simulation.

    Inputs:
        - water_source:
            The data for the :class:`conversion.WaterSource` to be processed.
        - availability:
            A :class:`pandas.DataFrame` giving the availability of the
            :class:`WaterSource` being considered.
        - generated_water_source_availability_directory:
            The directory in which to store the generated water-source availability
            profiles.
        - keyword:
            A `str` used for differentiating the different types of conventional source
            when printing or making log calls.
        - location:
            The location currently being considered.
        - logger:
            The logger to use for the run.
        - regenerate:
            Whether to force-regenerate the profiles.

    Outputs:
        - The interpolated daily availability profile for each water source.

    """

    logger.info(
        "Water-source availability process instantiated for %s water-source %s.",
        keyword,
        water_source.name,
    )

    daily_times_filename = f"{water_source.name}_daily_times.csv"
    filepath = os.path.join(
        generated_water_source_availability_directory, daily_times_filename
    )

    # If the file already exists, simply read in the data.
    if os.path.isfile(filepath) and not regenerate:
        with open(filepath, "r") as f:
            interpolated_daily_profile = pd.read_csv(f, header=None)
        logger.info(
            "Daily %s water-source availability profile for %s successfully read from "
            "file %s.",
            keyword,
            water_source.name,
            filepath,
        )

    else:
        logger.info(
            "Computing %s water-source availability profile for %s.",
            keyword,
            water_source.name,
        )
        interpolated_daily_profile = monthly_times_to_daily_times(
            availability,
            location.max_years,
        )
        logger.info(
            "Daily %s water-source availability profile for %s successfully computed.",
            keyword,
            water_source.name,
        )

        # Save this to the output file.
        with open(filepath, "w") as f:
            interpolated_daily_profile.to_csv(
                f, header=None, index=False, line_terminator=""  # type: ignore
            )
        logger.info(
            "Daily %s water-source availability profile for %s successfully saved to "
            "%s.",
            keyword,
            water_source.name,
            daily_times_filename,
        )

    return interpolated_daily_profile


def _process_water_soure_hourly_probability(
    water_source: WaterSource,
    *,
    daily_water_source_availability: pd.DataFrame,
    generated_water_source_availability_directory: str,
    keyword: str,
    logger: Logger,
    regenerate: bool,
) -> pd.DataFrame:
    """
    Calculate the probability that the water source is available at any given time.

    Inputs:
        - water_source:
            The data for the :class:`conversion.WaterSource` to be processed.
        - availability:
            A :class:`pandas.DataFrame` giving the availability of the
            :class:`WaterSource` being considered.
        - generated_water_source_availability_directory:
            The directory in which to store the generated water-source availability
            profiles.
        - keyword:
            A `str` used for differentiating the different types of conventional source
            when printing or making log calls.
        - logger:
            The logger to use for the run.
        - regenerate:
            Whether to force-regenerate the profiles.
        - years:
            The number of years for which the simulation is being run.

    Outputs:
        - The hourly availability of the conventional :class:`WaterSource` specified as
        a :class:`pandas.DataFrame`.

    """

    filename = f"{water_source.name}_availability.csv"
    filepath = os.path.join(generated_water_source_availability_directory, filename)

    # If the device hourly usage already exists, then read the data in from the file.
    if os.path.isfile(filepath) and not regenerate:
        with open(filepath, "r") as f:
            hourly_availability = pd.read_csv(f, header=None)
        logger.info(
            "Hourly %s water-source availability for %s successfully read from file: "
            "%s",
            keyword,
            water_source.name,
            filepath,
        )

    else:
        daily_water_source_availability = daily_water_source_availability.reset_index(
            drop=True
        )

        # Calculate the hourly-usage profile.
        logger.info("Calculating probability of %s availability", water_source.name)
        try:
            hourly_availability = pd.concat(
                [
                    daily_water_source_availability[column]
                    for column in daily_water_source_availability.columns
                ],
                axis=0,
            )
        except ValueError as e:
            logger.error(
                "%sError computing hourly %s water-source availability usage profile "
                "for %s: type error in variables: %s%s",
                BColours.fail,
                keyword,
                water_source.name,
                str(e),
                BColours.endc,
            )
            raise

        hourly_availability = hourly_availability.reset_index(drop=True)

        logger.info(
            "Hourly %s water-source availability profile for %s successfully "
            "calculated.",
            keyword,
            water_source.name,
        )

        # Save the hourly-usage profile.
        logger.info(
            "Saving hourly %s water-source availability profile for %s.",
            keyword,
            water_source.name,
        )

        with open(
            filepath,
            "w",
        ) as f:
            hourly_availability.to_csv(
                f, header=None, index=False, line_terminator=""  # type: ignore
            )

        logger.info(
            "Hourly %s water-source availability proifle for %s successfully saved to "
            "%s.",
            keyword,
            water_source.name,
            filename,
        )

    return hourly_availability


def get_lifetime_water_source_status(
    disable_tqdm: bool,
    generation_directory: str,
    keyword: str,
    location: Location,
    logger: Logger,
    regenerate: bool,
    water_source_times: Dict[WaterSource, pd.DataFrame],
) -> Dict[WaterSource, pd.DataFrame]:
    """
    Calculates, and saves, the water-source availability profiles of all input types.

    Inputs:
        - disable_tqdm:
            Whether to disable the tqdm progress bars (True) or display them (False).
        - generation_directory:
            The directory in which auto-generated files should be saved.
        - keyword:
            A `str` used for differentiating the different types of conventional source
            when printing or making log calls.
        - location:
            The :class:`Location` being considered.
        - logger:
            The logger to use for the run.
        - max_years:
            The maximum number of years for which the simulation should run.
        - regenerate:
            Whether to renerate (True) the profiles or use existing profiles if present
            (False).
        - water_source_timess:
            Water-source inputs information, read from the water-source inputs file and
            directory, stored as a mapping between :class:`WaterSource` instances,
            representing the conventional :class:`WaterSource`s available to the system,
            and :class:`pandas.DataFrame` instances giving their availability profiles.

    Outputs:
        - water_source_profiles:
            A dictionary mapping the water-source name to its associated maximum-output
            profile.

    """

    water_source_profiles: Dict[WaterSource, pd.DataFrame] = {}

    # Do not compute conventional profiles if there are none available.
    if len(water_source_times) == 0:
        tqdm.write(
            f"No conventional {keyword} water profiles defined "
            + "." * (35 - (len(keyword) + len(SKIPPING)))
            + f"    {SKIPPING}"
        )
        return water_source_profiles

    for source, availability in tqdm(
        water_source_times.items(),
        desc=f"conventional {keyword} water availability",
        disable=disable_tqdm,
        leave=True,
        unit="source",
    ):
        # Transform the profiles into daily profiles by interpolating across the months.
        interpolated_daily_profile = _process_water_source_availability(
            source,
            availability=availability,
            generated_water_source_availability_directory=os.path.join(
                generation_directory, "available_times"
            ),
            keyword=keyword,
            location=location,
            logger=logger,
            regenerate=regenerate,
        )

        # Compute the hourly water-source availability profile.
        hourly_availability = _process_water_soure_hourly_probability(
            source,
            daily_water_source_availability=interpolated_daily_profile,
            generated_water_source_availability_directory=os.path.join(
                generation_directory, "available_probability"
            ),
            keyword=keyword,
            logger=logger,
            regenerate=regenerate,
        )

        water_source_profiles[source] = (
            hourly_availability * source.maximum_output_capacity
        )

    return water_source_profiles
